{% load static %}

{% block favicons %}
    <link rel="icon" href="{% static 'ninja/favicon.svg' %}" type="image/svg+xml">
    <link rel="icon" href="{% static 'ninja/favicon.png' %}" type="image/png">
{% endblock %}